<?php

/**
 * Filter query for order report so that:
 * - we only consider data for the seller if the current user is a seller.
 * - if a seller, and the site uses custom completed order statuses, replace wc-completed with them in the query.
 *
 * @param array $query
 *
 * @return array
 */
function bt_filter_wc_order_report_query( $query ) {
	//error_log( 'FILTER ORDER REPORT QUERY: query = ' . print_r( $query, 1 ) );

	if ( bt_is_seller() ) {
		$query['where'] .= " AND posts.post_author = " . get_current_user_id() . " ";

		// handle custom order completed statuses if set
		if ( $custom_completed_statuses = bt_use_sellers( 'sellers_custom_order_completed_statuses' ) ) {
			// build string to swap in
			$status_string = '';
			foreach ( $custom_completed_statuses as $status ) {
				$status_string .= "'$status',";
			}
			$status_string = rtrim( $status_string, ',' );

			// replace wc-completed with string of custom statuses
			$query['where'] = str_replace( "'wc-completed'", $status_string, $query['where'] );
		}

		//error_log( 'FILTER ORDER REPORT QUERY: query ADJUSTED = ' . print_r( $query, 1 ) );
	}

	return $query;
}
add_filter( 'woocommerce_reports_get_order_report_query', 'bt_filter_wc_order_report_query', 9988, 1 );


/**
 * Filter query for any stock report so that:
 * - we only consider data for the seller if the current user is a seller.
 *
 * @param string $query
 *
 * @return string
 */
function bt_filter_wc_stock_report_query( $query ) {
	//error_log( 'FILTER STOCK REPORT QUERY: query = ' . $query );

	if ( bt_is_seller() ) {
		$query .= " AND posts.post_author = " . get_current_user_id() . " ";
		//error_log( 'FILTER STOCK REPORT QUERY: query ADJUSTED = ' . $query );
	}

	return $query;
}
add_filter( 'woocommerce_report_most_stocked_query_from', 'bt_filter_wc_stock_report_query', 9988, 1 );
add_filter( 'woocommerce_report_low_in_stock_query_from', 'bt_filter_wc_stock_report_query', 9988, 1 );
add_filter( 'woocommerce_report_out_of_stock_query_from', 'bt_filter_wc_stock_report_query', 9988, 1 );


/**
 * Filter available admin reports (before Woo 4.0 Analytics) so that:
 * - for sellers, we remove coupon, download, and customer list
 * - we add tips report if used
 * - we add Sellers report for admins and marketplace admins, when using sellers
 *
 * @param array $reports
 */
add_filter( 'woocommerce_admin_reports', function( $reports ) {
	//error_log( 'FILTER ADMIN REPORTS: reports = ' . print_r( $reports, 1 ) );

	// omit non-applicable reports for sellers
	if ( bt_is_seller() ) {
		unset( $reports['orders']['reports']['coupon_usage'] );
		unset( $reports['orders']['reports']['downloads'] );
		unset( $reports['customers']['reports']['customers'] );
		unset( $reports['customers']['reports']['customer_list'] );
		unset( $reports['subscriptions'] );

		// add custom customers report
		$reports['customers']['reports']['seller_customers'] = [
			'title' => 'Customer List',
			'description' => 'This report lists the customer for each order. Customers will be repeated if they ordered more than once.',
			'hide_title' => true,
			'callback' => 'bt_seller_customer_list',
		];
	}

	if ( bt_use_tips() ) {
		// add tips reports
		$reports['tips'] = [
			'title' => 'Tips',
			'reports' => [
				'tips_processing' => [
					'title' => 'Processing',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_tips_report_processing',
				],
				'tips_completed' => [
					'title' => 'Completed',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_tips_report_completed',
				],
				'tips_all' => [
					'title' => 'All',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_tips_report',
					/*'callback' => [
						'WC_Admin_Reports',
						'get_report',
					],*/
				],
			],
		];
	}

	if ( bt_use_sellers() && bt_can_admin_marketplace() ) {
		// CUSTOM SELLER REPORTS FOR ONLY ADMINS TO SEE
		$admin_reports = [];

		// add monitoring reports
		$admin_reports['monitoring'] = [
			'title' => 'Monitoring',
			'reports' => [
				'monitor_open_orders' => [
					'title' => 'Open Orders',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_monitoring_report_orders_open',
				],
				'monitor_order_issues' => [
					'title' => 'Order Issues',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_monitoring_report_orders',
				],
			],
		];

		// add payouts reports
		$admin_reports['payouts'] = [
			'title' => 'Payouts',
			'reports' => [
				'payouts_to_do' => [
					'title' => 'To Do',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_payouts_report_to_do',
				],
				'payouts_to_do_stripe' => [
					'title' => 'To Do - Stripe Only',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_payouts_report_to_do_stripe_payouts',
				],
				'payouts_to_do_manual' => [
					'title' => 'To Do - Manual Only',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_payouts_report_to_do_manual_payouts',
				],
				'payouts_audit_imbalance' => [
					'title' => 'Audit for Imbalance',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_payouts_report_audit_imbalance',
				],
				'payouts_historical' => [
					'title' => 'Historical',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_payouts_report_historical',
				],
			],
		];

		// add sellers reports
		$admin_reports['sellers'] = [
			'title' => 'Sellers',
			'reports' => [
				'sellers_approved' => [
					'title' => 'Approved',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_status_approved',
				],
				'sellers_manual' => [
					'title' => 'Manual Approval Needed',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_status_manual',
				],
				'sellers_stripe_payouts' => [
					'title' => 'Stripe Payouts',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_status_stripe_payouts',
				],
				'sellers_manual_payouts' => [
					'title' => 'Manual Payouts',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_status_manual_payouts',
				],
				'sellers_incomplete' => [
					'title' => 'Incomplete Setup',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_status_incomplete',
				],
				'sellers_status' => [
					'title' => 'All Sellers',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_status',
				],
				'sellers_external' => [
					'title' => 'External Shops',
					'description' => '',
					'hide_title' => true,
					'callback' => 'bt_sellers_report_external',
				],
			],
		];

		// add subscribers reports
		if ( class_exists( 'WC_Subscriptions' ) ) {
			$admin_reports['subscribers'] = [
				'title' => (
					bt_ecomm_option( 'subscriptions_as_membership' )
						? 'Members'
						: 'Subscribers'
				),
				'reports' => [
					'subscribers_active' => [
						'title' => 'Active',
						'description' => '',
						'hide_title' => true,
						'callback' => 'bt_subscribers_report_active',
					],
					'subscribers_inactive' => [
						'title' => 'Inactive',
						'description' => '',
						'hide_title' => true,
						'callback' => 'bt_subscribers_report_inactive',
					],
					'subscribers_all' => [
						'title' => 'All',
						'description' => '',
						'hide_title' => true,
						'callback' => 'bt_subscribers_report',
					],
				],
			];
		}

		// combine with all reports, but before them
		$reports = $admin_reports + $reports;

		//error_log( 'FILTER ADMIN REPORTS [END]: reports = ' . print_r(  $reports, 1 ) );
	}

	return $reports;
}, 9988, 1);


/**
 * Run the tips report for processing orders only
 */
function bt_tips_report_processing() {
	bt_tips_report( [
		'post_status' => ['wc-processing', 'wc-ready-pickup'],
	] );
}


/**
 * Run the tips report for completed orders only
 */
function bt_tips_report_completed() {
	bt_tips_report( [
		'post_status' => ['wc-completed', 'wc-completed-picked', 'wc-completed-shipped'],
	] );
}


/**
 * Run and echo tips report. By default includes both processing and completed orders.
 *
 * @param array $args Anything to override defaults
 */
function bt_tips_report( $args = [] ) {
	// define how many orders we'll show
	$max_rows = 1000;

	// default query arguments
	$defaults = [
		'post_type' => 'shop_order',
		'nopaging' => true,
		'posts_per_page' => $max_rows,
		'post_status' => [ 'wc-processing', 'wc-ready-pickup', 'wc-completed', 'wc-completed-picked', 'wc-completed-shipped' ],
		'orderby' => 'date',
		'order' => 'DESC',
	];

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// restrict to seller data if it's a seller
	if ( bt_is_seller() ) {
		$parsed_args['author'] = get_current_user_id();
	}

	// run query
	$orders = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $orders->posts ) ) {
		echo '<p><em>There aren\'t any orders to show for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>Order #</th>
					<th>Date Placed</th>
					<th>Status</th>
					<th class="number">Subtotal</th>
					<th class="number">Tip %</th>
					<th class="number">Tip $</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$total_orders = 0;
	$total_tips = 0;
	$date_current = '';

	// loop orders
	foreach ( $orders->posts as $order ) {
		// formal WC object
		$wc_order = new \WC_Order( $order );

		// count up, add subtotal, init tip vars
		$count++;
		$total_orders += $wc_order->get_subtotal();
		$tip_amount = 0;
		$tip_amount_html = '';
		$tip_percent_html = '';
		$date_is_new = false;

		// find our tip fee
		foreach ( $wc_order->get_fees() as $fee ) {
			if ( strpos( $fee->get_name(), 'Tip' ) !== false ) {
				// get amount
				$tip_amount = $fee->get_total();
				// build amount and percentage html
				$tip_amount_html = '$' . number_format( $tip_amount, 2 );
				$tip_percent_html = trim( str_replace( 'Tip', '', $fee->get_name() ) );
				// add total tips
				$total_tips += floatval( $tip_amount );
				break;
			}
		}

		// date formatting and checking
		$date = wp_date( 'Y-m-d g:ia', strtotime( $wc_order->get_date_created() ) );
		$date_parts = explode( ' ', $date );
		if ( ! empty( $date_current ) && $date_current != $date_parts[0] ) {
			// date has been set but is new; update it
			$date_is_new = true;
			$date_current = $date_parts[0];
		} else if ( empty( $date_current ) ) {
			// date hasn't been set; this is first time so init to the first order's date
			$date_current = $date_parts[0];
		}

		// output row for order
		echo '
			<tr' . ( $date_is_new ? ' class="date-new"' : '' ) . '>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $order->ID . '&action=edit">' . $order->ID . '</a>
				</td>
				<td>' . $date . '</td>
				<td>' . ucfirst( $wc_order->get_status() ) . '</td>
				<td class="number">$' . number_format( $wc_order->get_subtotal(), 2) . '</td>
				<td class="number">' . $tip_percent_html . '</td>
				<td class="number">' . $tip_amount_html . '</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="3">Total: ' . $count . ' Order' . ( $count == 1 ? '' : 's' ) . '</td>
					<td class="number">$' . number_format( $total_orders, 2 ) . '</td>
					<td class="number"> ' . number_format( 100 * ( $total_tips / $total_orders ), 2 ) . '%</td>
					<td class="number">$' . number_format( $total_tips, 2 ) . '</td>
				</tr>
			</tfoot>
		</table>';

	// warn if we only showed up to the max
	if ( count( $orders->posts ) > $max_rows ) {
		echo '<p><em>Limited report to the ' . $max_rows . ' most recent orders.</em></p>';
	}
}


/**
 * Run the sellers report for manual approval needed
 */
function bt_sellers_report_status_manual() {
	bt_sellers_report_status( [], 'manual' );
}


/**
 * Run the sellers report for incomplete
 */
function bt_sellers_report_status_incomplete() {
	bt_sellers_report_status( [], 'incomplete' );
}


/**
 * Run the sellers report for approved
 */
function bt_sellers_report_status_approved() {
	bt_sellers_report_status( [], 'approved' );
}


/**
 * Run the sellers report for stripe payouts
 */
function bt_sellers_report_status_stripe_payouts() {
	bt_sellers_report_status( [], 'stripe_payouts' );
}


/**
 * Run the sellers report for manual payouts
 */
function bt_sellers_report_status_manual_payouts() {
	bt_sellers_report_status( [], 'manual_payouts' );
}


/**
 * Run and echo sellers status report.
 *
 * @param array $args Anything to override defaults
 * @param string $mode
 */
function bt_sellers_report_status( $args = [], $mode = '' ) {
	// define how many orders we'll show
	//$max_rows = 1000;

	// default query arguments
	$defaults = [
		'post_type' => bt_use_sellers( 'seller_post_type' ),
		'nopaging' => true,
		'posts_per_page' => -1, //$max_rows,
		'post_status' => [ 'published' ],
		'orderby' => 'title',
		'order' => 'ASC',
		'meta_query' => [
			[
				'key' => 'sell_on_this_marketplace',
				'value' => '0',
				'compare' => '!=',
			],
		],
	];

	// exclude any sellers set by bt-sellers theme support
	bt_exclude_sellers_from_reports_arg( $defaults );

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$sellers = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $sellers->posts ) ) {
		echo '<p><em>There aren\'t any sellers to show for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>ID</th>
					<th>Name</th>
					<th>Type</th>
					<th>User</th>
					<th class="flag">Contract</th>
					<th class="flag">Tax Reg</th>
					<th class="flag">Payouts</th>
					<th>Sell Mode</th>
					<th>Products</th>
					<th>Orders</th>
					<th>Email</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$count_curbside = 0;
	$count_usps = 0;
	$count_both = 0;
	$count_products_publish = 0;
	$count_products_draft = 0;
	$count_orders_completed = 0;
	$count_orders_processing = 0;
	$count_orders_pending = 0;
	$count_orders_refunded = 0;
	$count_orders_trash = 0;
	$all_emails = [];


	// loop sellers
	foreach ( $sellers->posts as $seller ) {
		// set user from author
		$user_id = $seller->post_author;

		// init requirements flags
		// - main approved flag
		$approved = bt_seller_is_approved( $seller );
		// - step 1
		$contract = bt_seller_has_agreed_to_contract( $user_id ) ? true : false; // returns array of data if true
		// - step 2
		$tax_registration = bt_seller_has_sales_tax_registration( $user_id ); // false if no, array if true
		$tax_registration_file_url = '';
		if ( ! empty( $tax_registration ) ) {
			// get the file url and set our var to a true bool
			$tax_registration_file_url = $tax_registration['url'];
			$tax_registration = true;
		}
		// - step 3
		$manual_payouts = bt_seller_has_agreed_to_manual_payouts( $user_id );
		$stripe_payouts = bt_get_user_stripe_account_id( $user_id ) ? true : false; // returns string for account id if true
		// reset the manual payouts flag if stripe is also set, so those who used to have manual but changed to stripe aren't considered manual anywhere
		if ( $manual_payouts && $stripe_payouts ) {
			$manual_payouts = false;
		}

		// check for full requirements and if manual approval is needed
		$all_requirements = $contract && $tax_registration && ( $manual_payouts || $stripe_payouts );
		$manual_approval_required = $all_requirements && ! $approved;

		// set an alert class for the post name based on approval status
		$approval_class = 'warning';
		if ( $approved ) {
			$approval_class = 'success';
		} else if ( $manual_approval_required ) {
			$approval_class = 'info';
		}

		// bail if we're using a certain mode and this record doesn't fit it
		if ( $mode ) {
			if ( $mode == 'manual' ) {
				// manual approval required
				if ( ! $manual_approval_required ) {
					// exclude record from this mode
					continue;
				}
			} else if ( $mode == 'incomplete' ) {
				// sell-on-this-marketplace is on, but some steps are incomplete
				if ( $approved ) {
					// exclude record from this mode
					continue;
				}
			} else if ( $mode == 'approved' ) {
				// approved only
				if ( ! $approved ) {
					// exclude record from this mode
					continue;
				}
			} else if ( $mode == 'stripe_payouts' ) {
				// only those set for stripe payouts
				if ( ! $stripe_payouts ) {
					// exclude record from this mode
					continue;
				}
			} else if ( $mode == 'manual_payouts' ) {
				// only those set for manual payouts
				if ( ! $manual_payouts ) {
					// exclude record from this mode
					continue;
				}
			}
		}

		// count up
		$count++;

		// shipping methods
		$does_curbside = bt_seller_does_curbside( $seller );
		$does_usps = bt_seller_does_usps( $seller );
		$does_both = $does_curbside && $does_usps;
		if ( $does_both ) {
			$sell_mode = 'Both';
			$count_both++;
		} else if ( $does_usps ) {
			$sell_mode = 'Shipping';
			$count_usps++;
		} else {
			$sell_mode = 'Curbside';
			$count_curbside++;
		}

		// payout info
		$payout = 'No';
		$payout_class = 'warning';
		if ( $stripe_payouts ) {
			$payout = 'Stripe';
			$payout_class = 'success';
		} else if ( $manual_payouts ) {
			$payout = 'Manual';
			$payout_class = 'info';
		}

		// seller email
		$email = get_the_author_meta( 'user_email', $user_id  );
		$all_emails[] = $email;

		// product status counts
		$seller_products_publish = bt_get_user_post_count( $user_id, 'product', "'publish'" );
		$seller_products_draft = bt_get_user_post_count( $user_id, 'product', "'draft','trash'" );
		$count_products_publish += $seller_products_publish;
		$count_products_draft += $seller_products_draft;

		// order status counts
		$seller_orders_completed = bt_get_user_post_count( $user_id, 'shop_order', "'wc-completed','wc-completed-picked','wc-completed-shipped'" );
		$seller_orders_processing = bt_get_user_post_count( $user_id, 'shop_order', "'wc-processing','wc-ready-pickup'" );
		$seller_orders_pending = bt_get_user_post_count( $user_id, 'shop_order', "'wc-pending','wc-on-hold'" );
		$seller_orders_refunded = bt_get_user_post_count( $user_id, 'shop_order', "'wc-cancelled','wc-refunded'" );
		$seller_orders_trash = bt_get_user_post_count( $user_id, 'shop_order', "'trash'" );
		$count_orders_completed += $seller_orders_completed;
		$count_orders_processing += $seller_orders_processing;
		$count_orders_pending += $seller_orders_pending;
		$count_orders_refunded += $seller_orders_refunded;
		$count_orders_trash += $seller_orders_trash;

		// output row for seller
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $seller->ID . '&action=edit">' . $seller->ID . '</a>
				</td>
				<td>
					<div class="alert alert-' . $approval_class . '">
						<a href="' . get_the_permalink( $seller ) . '">' . $seller->post_title . '</a>
					</div> 
				</td>
				<td>' . ucfirst( $seller->post_type ) . '</td>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user_id . '">User ' . $user_id . '</a>
				</td>
				<td class="flag">
					' . ( $contract
						? '<div class="alert alert-success">Yes</div>'
						: '<div class="alert alert-warning">No</div>' ) . '
				</td>
				<td class="flag">
					' . ( $tax_registration
						? '<div class="alert alert-success"><a href="' . $tax_registration_file_url . '" target="_blank">Yes</a></div>'
						: '<div class="alert alert-warning">No</div>' ) . '
				</td>
				<td class="flag">
					<div class="alert alert-' . $payout_class . '">' . $payout . '</div>
				</td>
				<td>' . $sell_mode . '</td>
				<td>
					<a href="' . get_admin_url() . 'edit.php?post_type=product&author=' . $user_id . '">Products</a>
					<span class="alert alert-success" title="Published">' . $seller_products_publish . '</span>
					<span class="alert alert-warning" title="Draft or Trash">' . $seller_products_draft . '</span>
				</td>
				<td>
					<a href="' . get_admin_url() . 'edit.php?post_type=shop_order&author=' . $user_id . '">Orders</a>
					<span class="alert alert-info" title="Completed">' . $seller_orders_completed . '</span>
					<span class="alert alert-success" title="Processing">' . $seller_orders_processing . '</span>
					<span class="alert alert-warning" title="Pending or On Hold">' . $seller_orders_pending . '</span>
					<span class="alert alert-danger" title="Cancelled or Refunded">' . $seller_orders_refunded . '</span>
					<span class="alert alert-danger" title="Trash">' . $seller_orders_trash . '</span>
				</td>
				<td>
					<a href="mailto:' . $email . '">' . $email . '</a>
				</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="8">
						Total: ' . $count . ' Seller' . ( $count == 1 ? '' : 's' ) . '
						&mdash;
						Curbside only: ' . $count_curbside . '
						&mdash;
						Shipping only: ' . $count_usps . '
						&mdash;
						Both: ' . $count_both . '
					</td>
					<td>
						<span class="alert alert-success" title="Published">' . $count_products_publish . '</span>
						<span class="alert alert-warning" title="Draft or Trash">' . $count_products_draft . '</span>
					</td>
					<td>
						<span class="alert alert-info" title="Completed">' . $count_orders_completed . '</span>
						<span class="alert alert-success" title="Processing">' . $count_orders_processing . '</span>
						<span class="alert alert-warning" title="Pending or On Hold">' . $count_orders_pending . '</span>
						<span class="alert alert-danger" title="Cancelled or Refunded">' . $count_orders_refunded . '</span>
						<span class="alert alert-danger" title="Trash">' . $count_orders_trash . '</span>
					</td>
					<td></td>
				</tr>
			</tfoot>
		</table>
		<textarea rows="8" style="margin-top: 2rem; width: 100%;">' . implode( ', ', $all_emails ) . '</textarea>
		';

	// warn if we only showed up to the max
	/*
	if ( count( $sellers->posts ) > $max_rows ) {
		echo '<p><em>Limited report to the first ' . $max_rows . ' sellers.</em></p>';
	}
	*/
}


/**
 * Run and echo external sellers report.
 *
 * @param array $args Anything to override defaults
 */
function bt_sellers_report_external( $args = [] ) {
	// define how many orders we'll show
	//$max_rows = 1000;

	// default query arguments
	$defaults = [
		'post_type' => bt_use_sellers( 'seller_post_type' ),
		'nopaging' => true,
		'posts_per_page' => -1, //$max_rows,
		'post_status' => [ 'published' ],
		'orderby' => 'title',
		'order' => 'ASC',
		'meta_query' => [
			'relation' => 'AND',
			[
				'relation' => 'OR',
				[
					'key' => 'sell_on_this_marketplace',
					'value' => '0',
					'compare' => '=',
				],
				[
					'key' => 'sell_on_this_marketplace',
					'compare' => 'NOT EXISTS',
				],
			],
			[
				'key' => 'external_ordering_url',
				'value' => '',
				'compare' => '!=',
			],
		],
	];

	// exclude any sellers set by bt-sellers theme support
	bt_exclude_sellers_from_reports_arg( $defaults );

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$sellers = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $sellers->posts ) ) {
		echo '<p><em>There aren\'t any sellers to show for this report.</em></p>';
		return;
	}

	// init var to hold all emails for output in textarea
	$all_emails = [];

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>ID</th>
					<th>Name</th>
					<th>Type</th>
					<th>User</th>
					<th>Email</th>
					<th>External Shop</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$count_curbside = 0;
	$count_usps = 0;
	$count_both = 0;

	// loop sellers
	foreach ( $sellers->posts as $seller ) {
		// set user from author
		$user_id = $seller->post_author;

		// count up
		$count++;

		// seller email
		$email = get_the_author_meta( 'user_email', $user_id  );
		$all_emails[] = $email;

		// external ordering url
		$external_url = get_field( 'external_ordering_url', $seller  );

		// output row for seller
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $seller->ID . '&action=edit">' . $seller->ID . '</a>
				</td>
				<td>
					<a href="' . get_the_permalink( $seller ) . '">' . $seller->post_title . '</a>
				</td>
				<td>' . ucfirst( $seller->post_type ) . '</td>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user_id . '">User ' . $user_id . '</a>
				</td>
				<td>
					<a href="mailto:' . $email . '">' . $email . '</a>
				</td>
				<td>
					<a href="' . $external_url . '">' . $external_url . '</a>
				</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="8">
						Total: ' . $count . ' Seller' . ( $count == 1 ? '' : 's' ) . '
					</td>
				</tr>
			</tfoot>
		</table>
		<textarea rows="8" style="margin-top: 2rem; width: 100%;">' . implode( ', ', $all_emails ) . '</textarea>
	';

	// warn if we only showed up to the max
	/*
	if ( count( $sellers->posts ) > $max_rows ) {
		echo '<p><em>Limited report to the first ' . $max_rows . ' sellers.</em></p>';
	}
	*/
}


/**
 * Run the subscribers report for active subscribers
 */
function bt_subscribers_report_active() {
	bt_subscribers_report( [], 'active' );
}


/**
 * Run the subscribers report for inactive subscribers
 */
function bt_subscribers_report_inactive() {
	bt_subscribers_report( [], 'inactive' );
}


/**
 * Run and echo subscribers report.
 *
 * @param array $args Anything to override defaults
 * @param string $report_status If 'active' get users with active subscriptions. If 'inactive' get users without.
 *      Otherwise, get all.
 */
function bt_subscribers_report( $args = [], $report_status = '' ) {
	// define how many orders we'll show
	//$max_rows = 1000;

	// default query arguments
	if ( ! $roles = bt_ecomm_option( 'member_roles' ) ) {
		$roles = ['subscriber'];
	}
	$defaults = [
		'role__in' => $roles,
		'orderby' => 'display_name',
		'order' => 'ASC',
	];

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$users = new \WP_User_Query( $parsed_args );
	//echo '<pre>' . print_r( $users, 1 ) . '</pre>';

	// if none, message and exit
	if ( empty( $users ) ) {
		echo '<p><em>There aren\'t any subscribers to show for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>ID</th>
					<th>Name</th>
					<th>Email</th>
					' . ( bt_use_sellers() ? '<th>Post</th>' : '' ) . '
					<th class="flag">Active</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$count_active = 0;
	$count_inactive = 0;
	/*
	$count_products_publish = 0;
	$count_products_draft = 0;
	$count_orders_completed = 0;
	$count_orders_processing = 0;
	$count_orders_pending = 0;
	$count_orders_refunded = 0;
	$count_orders_trash = 0;
	*/
	$all_emails = [];


	// loop sellers
	foreach ( $users->results as $user ) {
		//error_log( __METHOD__ . ' -- $user = ' . print_r( $user, true ) );
		//continue;

		// check for active subscription status
		$subscription_active = false;
		if ( wcs_user_has_subscription( $user->ID, '', 'active' ) ) {
			$subscription_active = true;
		}

		// include this user in report?
		if ( $report_status == 'active' && ! $subscription_active ) {
			// report is active users, this user doesn't have a subscription, so skip them
			continue;
		} else if ( $report_status == 'inactive' && $subscription_active ) {
			// report is inactive users, this user has a subscription, so skip them
			continue;
		}

		// count up
		$count++;
		if ( $subscription_active ) {
			$count_active++;
		} else {
			$count_inactive++;
		}

		// track email
		$all_emails[] = $user->user_email;

		// get corresponding post
		$post = null;
		if ( bt_use_sellers() ) {
			$post = bt_get_seller_post( $user->ID );
		}

		// product status counts
		/*
		$seller_products_publish = bt_get_user_post_count( $user_id, 'product', "'publish'" );
		$seller_products_draft = bt_get_user_post_count( $user_id, 'product', "'draft','trash'" );
		$count_products_publish += $seller_products_publish;
		$count_products_draft += $seller_products_draft;

		// order status counts
		$seller_orders_completed = bt_get_user_post_count( $user_id, 'shop_order', "'wc-completed','wc-completed-picked','wc-completed-shipped'" );
		$seller_orders_processing = bt_get_user_post_count( $user_id, 'shop_order', "'wc-processing','wc-ready-pickup'" );
		$seller_orders_pending = bt_get_user_post_count( $user_id, 'shop_order', "'wc-pending','wc-on-hold'" );
		$seller_orders_refunded = bt_get_user_post_count( $user_id, 'shop_order', "'wc-cancelled','wc-refunded'" );
		$seller_orders_trash = bt_get_user_post_count( $user_id, 'shop_order', "'trash'" );
		$count_orders_completed += $seller_orders_completed;
		$count_orders_processing += $seller_orders_processing;
		$count_orders_pending += $seller_orders_pending;
		$count_orders_refunded += $seller_orders_refunded;
		$count_orders_trash += $seller_orders_trash;
		*/

		// output row for subscriber
		/*
		WP_User
        (
            [ID] => 258
            [user_login] => tom-riggle
            [user_pass] => ...
            [user_nicename] => tom-riggle
            [user_email] => grainbeltart@gmail.com
            [user_url] =>
            [user_registered] => 2020-04-26 00:13:32
            [user_activation_key] =>
            [user_status] => 0
            [display_name] => 1 Abattoir Studios-Tom Riggle
        )
		*/
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user->ID . '">User ' . $user->ID . '</a>
				</td>
				<td>
					' . $user->display_name . '
				</td>
				<td>
					<a href="mailto:' . $user->user_email . '">' . $user->user_email . '</a>
				</td>
				' . ( bt_use_sellers()
					? '<td>' . ucfirst( get_post_type( $post ) ) . ' ' . $post->ID . ': '
					    . '<a href="' . get_admin_url() . 'post.php?post=' . $post->ID . '&action=edit" target="_blank">Edit</a>'
					    . ' &middot; <a href="' . get_the_permalink( $post ) . '" target="_blank">View</a>'
					  . '</td>'
					: ''
				) . '
				<td class="flag">
					' . ( $subscription_active
						? '<div class="alert alert-success">Yes</div>'
						: '<div class="alert alert-warning">No</div>'
					) . '
				</td>
			</tr>
		';
	}

	// finish table
	$count_html = 'Total: ' . $count . ' Subscriber' . ( $count == 1 ? '' : 's' );
	if ( $count_active ) {
		$count_html.= ' &mdash; Active: ' . $count_active;
	}
	if ( $count_inactive ) {
		$count_html.= ' &mdash; Inctive: ' . $count_inactive;
	}
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="' . ( bt_use_sellers() ? '5' : '4' ) . '">
						 ' . $count_html . '
					</td>
				</tr>
			</tfoot>
		</table>
		<textarea rows="8" style="margin-top: 2rem; width: 100%;">' . implode( ', ', $all_emails ) . '</textarea>
		';

	// warn if we only showed up to the max
	/*
	if ( count( $sellers->posts ) > $max_rows ) {
		echo '<p><em>Limited report to the first ' . $max_rows . ' sellers.</em></p>';
	}
	*/
}


/**
 * Run and echo monitoring report for orders. Will show/flag orders which:
 * - Are in the trash
 * - Have multiple sellers
 * - Have a shipping address different from billing
 * - Are still processing after 4 days
 *
 * @param array $args Anything to override defaults
 */
function bt_monitoring_report_orders( $args = [] ) {
	// define how many orders we'll show
	$max_rows = 1000;

	// default query arguments
	$defaults = [
		'post_type' => 'shop_order',
		'nopaging' => true,
		'posts_per_page' => $max_rows,
		'post_status' => [ 'any', 'trash' ],
		'orderby' => 'date',
		'order' => 'DESC',
	];

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$orders = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $orders->posts ) ) {
		echo '<p><em>There aren\'t any orders to show for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>ID</th>
					<th>Placed</th>
					<th>Status</th>
					<th>Seller Post</th>
					<th>Seller User</th>
					<th>Customer</th>
					<th>Issue</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$processing_time_limit = 4 /* DAYS */ * 60 * 60 * 24;

	// loop orders
	foreach ( $orders->posts as $order ) {
		// set user from author
		$user_id = $order->post_author;

		// seller email
		$email = get_the_author_meta( 'user_email', $user_id  );

		// order object
		$order_object = wc_get_order( $order );

		// look for issues
		$issue = '';
		$issue_status = 'success';
		if ( $order->post_status == 'trash' ) {
			$issue = 'In trash';
			$issue_status = 'danger';

		} else if ( $order->post_status == 'wc-on-hold' ) {
			$issue = 'On hold';
			$issue_status = 'danger';

		} else if ( $order->post_status == 'wc-refunded' ) {
			$issue = 'Refunded';
			$issue_status = 'warning';

		} else if ( bt_can_admin_marketplace( $order->post_author ) ) {
			$issue = 'Admin (ID:' . $order->post_author . ') is owner';
			$issue_status = 'danger';

		} else if ( bt_order_has_multiple_sellers( $order_object ) ) {
			$issue = 'Multiple sellers';
			$issue_status = 'danger';

		} else if ( ! bt_order_shipping_matches_billing( $order_object ) && $order->post_status == 'wc-processing' ) {
			$issue = 'Ship-to address differs from billing';
			$issue_status = 'warning';

		} else if (
			$order_object->get_status() == 'processing'
			&& ( $order_object->get_date_created()->format( 'U' ) < ( wp_date( 'U' ) - $processing_time_limit ) )
		) {
			$issue = 'Incomplete after many days';
			$issue_status = 'warning';

		} else {
			// order is ok, move on to the next
			continue;
		}

		// get seller for order
		$seller = bt_get_order_seller_data( $order->ID );

		// count up
		$count++;

		// output row for order
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $order->ID . '&action=edit">' . $order->ID . '</a>
				</td>
				<td>' . $order_object->get_date_created()->format( 'n/j g:ia' ) . '</td>
				<td>' . ucfirst( $order_object->get_status() ) . '</td>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $seller->post->ID . '&action=edit">' . $seller->post->post_title . '</a>
				</td>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user_id . '">' . bt_seller_contract_name( $user_id ) . '</a>
				</td>
				<td>
					' . $order_object->get_billing_first_name() . ' ' . $order_object->get_billing_last_name() . '
				</td>
				<td>
					<div class="alert alert-' . $issue_status . '">' . $issue . '</div>
				</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="7">
						Total: ' . $count . ' Order' . ( $count == 1 ? '' : 's' ) . '
					</td>
				</tr>
			</tfoot>
		</table>';

	// warn if we only showed up to the max
	if ( count( $orders->posts ) > $max_rows ) {
		echo '<p><em>Limited report to the first ' . $max_rows . ' orders.</em></p>';
	}
}


/**
 * Run and echo monitoring report for open orders. Will show sellers which have orders in "Processing" status
 * and will flag how old the oldest is.
 *
 * @param array $args Anything to override defaults
 */
function bt_monitoring_report_orders_open( $args = [] ) {
	// default query arguments
	$defaults = [
		'post_type' => bt_use_sellers( 'seller_post_type' ),
		'nopaging' => true,
		'posts_per_page' => -1, //$max_rows,
		'post_status' => [ 'any', 'trash' ],
		'orderby' => 'title',
		'order' => 'ASC',
		/*'meta_query' => [
			[
				'key' => 'sell_on_this_marketplace',
				'value' => '0',
				'compare' => '!=',
			],
		],*/
	];

	// exclude any sellers set by bt-sellers theme support
	bt_exclude_sellers_from_reports_arg( $defaults );

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$sellers = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $sellers->posts ) ) {
		echo '<p><em>There aren\'t any sellers to show for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>ID</th>
					<th>Name</th>
					<th>Type</th>
					<th>User</th>
					<th class="flag">Approved</th>
					<th>Sell Mode</th>
					<th>Orders</th>
					<th class="flag">Days Old</th>
					<th>Email</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$count_orders_completed = 0;
	$count_orders_processing = 0;
	$all_emails = [];

	// loop sellers
	foreach ( $sellers->posts as $seller ) {
		// set user from author
		$user_id = $seller->post_author;

		// init requirements flags
		// - main approved flag
		$approved = bt_seller_is_approved( $seller );

		// set an alert class for the post name based on approval status
		$approval_class = 'warning';
		if ( $approved ) {
			$approval_class = 'success';
		}

		// shipping methods
		$does_curbside = bt_seller_does_curbside( $seller );
		$does_usps = bt_seller_does_usps( $seller );
		$does_both = $does_curbside && $does_usps;
		if ( $does_both ) {
			$sell_mode = 'Both';
		} else if ( $does_usps ) {
			$sell_mode = 'Shipping';
		} else {
			$sell_mode = 'Curbside';
		}

		// order status counts
		$seller_orders_completed = bt_get_user_post_count( $user_id, 'shop_order', "'wc-completed','wc-completed-picked','wc-completed-shipped'" );
		$seller_orders_processing = bt_get_user_post_count( $user_id, 'shop_order', "'wc-processing','wc-ready-pickup'" );
		$count_orders_completed += $seller_orders_completed;
		$count_orders_processing += $seller_orders_processing;

		// bail if no processing orders
		if ( $seller_orders_processing < 1 ) {
			continue;
		}

		// count up
		$count++;

		// seller email
		$email = get_the_author_meta( 'user_email', $user_id  );
		$all_emails[] = $email;

		// oldest open order date
		$oldest_open_date = bt_get_seller_oldest_open_order_date( $user_id );
		$age_seconds = wp_date( 'U' ) - strtotime( $oldest_open_date );
		$age_days = number_format( $age_seconds / ( 60 * 60 * 24 ), 1 );
		$age_alert_type = 'success';
		if ( $age_days >= 4 ) {
			$age_alert_type = 'danger';
		} else if ( $age_days >= 3 ) {
			$age_alert_type = 'warning';
		}

		// output row for seller
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $seller->ID . '&action=edit">' . $seller->ID . '</a>
				</td>
				<td>
					<div class="alert alert-' . $approval_class . '">
						<a href="' . get_the_permalink( $seller ) . '">' . $seller->post_title . '</a>
					</div> 
				</td>
				<td>' . ucfirst( $seller->post_type ) . '</td>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user_id . '">User ' . $user_id . '</a>
				</td>
				<td class="flag">
					' . ( $approved
						? '<div class="alert alert-success">Yes</div>'
						: '<div class="alert alert-warning">No</div>' ) . '
				</td>
				<td>' . $sell_mode . '</td>
				<td>
					<a href="' . get_admin_url() . 'edit.php?post_type=shop_order&author=' . $user_id . '">Orders</a>
					<span class="alert alert-info" title="Completed">' . $seller_orders_completed . '</span>
					<span class="alert alert-success" title="Processing">' . $seller_orders_processing . '</span>
				</td>
				<td class="flag">
					<div class="alert alert-' . $age_alert_type . '">' . $age_days . '</div>
				</td>
				<td>
					<a href="mailto:' . $email . '">' . $email . '</a>
				</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="6">
						Total: ' . $count . ' Seller' . ( $count == 1 ? '' : 's' ) . '
					</td>
					<td>
						<span class="alert alert-info" title="Completed">' . $count_orders_completed . '</span>
						<span class="alert alert-success" title="Processing">' . $count_orders_processing . '</span>
					</td>
					<td></td>
					<td></td>
				</tr>
			</tfoot>
		</table>
		<textarea rows="8" style="margin-top: 2rem; width: 100%;">' . implode( ', ', $all_emails ) . '</textarea>
		';
}


/**
 * Run the payouts report for stripe payouts
 */
function bt_payouts_report_to_do_stripe_payouts() {
	bt_payouts_report_to_do( [], 'stripe_payouts' );
}


/**
 * Run the payouts report for manual payouts
 */
function bt_payouts_report_to_do_manual_payouts() {
	bt_payouts_report_to_do( [], 'manual_payouts' );
}


/**
 * Run and echo report of sellers which have payouts to do, which includes:
 * - Sellers with completed orders for a week which can be paid out and hasn't yet.
 *
 * Only counts orders in payable weeks (last week if Wednesday or later; else the week before last).
 *
 * NOTE: This report will miss any orders from a week that already was paid but were completed after the payout.
 * Typically we'd see/handle it with a subsequent payout, but if there aren't any subsequent orders for that seller,
 * we wouldn't know about it.
 * ==> USE THE "AUDIT FOR IMBALANCE" REPORT TO SEE ANYTHING HIDDEN LIKE THIS.
 * Another potential solution:
 * - When running a payout, store the number of orders it accounts for in DB. (we'll only be able to report on sales since this feature)
 * - Create a "Monitoring" report:
 *   - Loop all sellers
 *     - Get bt_get_user_post_count() count of everything completed SINCE we started tracking order counts in payout DB.
 *       - bail if no sales since then
 *     - Loop payouts made and count up total orders paid
 *     - Compares the count, and if there's a difference, output a row in the report.
 *
 * @param array $args Anything to override defaults
 *
 * @param string $mode
 *
 * @throws Exception
 */
function bt_payouts_report_to_do( $args = [], $mode = '' ) {

	// get all sellers that sell on this marketplace
	$defaults = [
		'post_type' => bt_use_sellers( 'seller_post_type' ),
		'nopaging' => true,
		'posts_per_page' => -1, //$max_rows,
		'post_status' => [ 'published' ],
		'orderby' => 'title',
		'order' => 'ASC',
		'meta_query' => [
			[
				'key' => 'sell_on_this_marketplace',
				'value' => '0',
				'compare' => '!=',
			],
		],
	];

	// exclude any sellers set by bt-sellers theme support
	bt_exclude_sellers_from_reports_arg( $defaults );

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$sellers = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $sellers->posts ) ) {
		echo '<p><em>There aren\'t any sellers to show for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>Seller Post</th>
					<th>User</th>
					<th>Email</th>
					<th>Method</th>
					<th class="flag">Last Payout</th>
					<th>Orders to Pay Out</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count = 0;
	$count_orders_to_pay_out = 0;

	// loop sellers
	foreach ( $sellers->posts as $seller ) {
		// set user from author
		$user_id = $seller->post_author;

		// init payout vars, get completed payouts
		$last_payout = 'Never';
		$since_last_payout = '';
		$seller_payouts_completed = bt_seller_payouts( $user_id );

		// if any payouts have happened, get last date for display,
		// and calculate date since then for which to get order count
		if ( ! empty( $seller_payouts_completed ) ) {
			$last_payout = current( $seller_payouts_completed )->date_paid;
			$last_payout_year_week = key( $seller_payouts_completed );
			// starting with a datetime of the WEEK which was last paid out, add 7 days to get the NEXT WEEK's starting point
			$since_last_payout_datetime = new \DateTime( $last_payout_year_week, new \DateTimeZone( get_option( 'timezone_string' ) ) );
			$since_last_payout_datetime->add( new \DateInterval( 'P7D' ) );
			// format for query
			$since_last_payout = $since_last_payout_datetime->format( 'Y-m-d H:i:s' );
		}

		// determine date for which orders must be OLDER THAN in order to pay out
		$payout_orders_through = bt_payout_orders_through();

		// order status counts
		// this is completed SINCE the start of eligible payout weeks, THROUGH the most recent eligible payout week
		$seller_orders_to_pay_out = bt_get_user_post_count(
			$user_id,
			'shop_order',
			"'wc-completed','wc-completed-picked','wc-completed-shipped'",
			$since_last_payout, // midnight of Monday following the most recent payout week, if any
			$payout_orders_through // midnight of Monday of last week if before Wednesday, else Monday of this week
		);

		// skip sellers without orders to pay out
		if ( $seller_orders_to_pay_out == 0 ) {
			continue;
		}

		// init seller info needed
		$manual_payouts = bt_seller_has_agreed_to_manual_payouts( $user_id );
		$stripe_payouts = bt_get_user_stripe_account_id( $user_id ) ? true : false; // returns string for account id if true
		// reset the manual payouts flag if stripe is also set, so those who used to have manual but changed to stripe aren't considered manual anywhere
		if ( $manual_payouts && $stripe_payouts ) {
			$manual_payouts = false;
		}

		// bail if we're using a certain mode and this record doesn't fit it
		if ( $mode ) {
			if ( $mode == 'stripe_payouts' ) {
				// only those set for stripe payouts
				if ( ! $stripe_payouts ) {
					// exclude record from this mode
					continue;
				}
			} else if ( $mode == 'manual_payouts' ) {
				// only those set for manual payouts
				if ( ! $manual_payouts ) {
					// exclude record from this mode
					continue;
				}
			}
		}

		// count up
		$count++;
		$count_orders_to_pay_out += $seller_orders_to_pay_out;

		// payouts type and alert class
		if ( $stripe_payouts ) {
			$payout_method = 'Stripe';
			$payout_method_alert_class = 'success';
		} else if ( $manual_payouts ) {
			$payout_method = 'Manual';
			$payout_method_alert_class = 'info';
		} else {
			$payout_method = 'None!';
			$payout_method_alert_class = 'danger';
		}

		// seller email
		$email = get_the_author_meta( 'user_email', $user_id  );

		// output row for seller
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $seller->ID . '&action=edit">' . $seller->post_title . '</a>
					(' . ucfirst( $seller->post_type ) . ')
				</td>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user_id . '">User ' . $user_id . '</a>
				</td>
				<td>
					<a href="mailto:' . $email . '">' . $email . '</a>
				</td>
				<td class="flag">
					<div class="alert alert-' . $payout_method_alert_class . '">
						' . $payout_method . '
					</div>
				</td>
				<td>' . $last_payout . '</td>
				<td>
					<a href="' . get_admin_url() . 'edit.php?post_type=shop_order&author=' . $user_id . '">Orders</a>
					<span class="alert alert-info" title="Completed">' . $seller_orders_to_pay_out . '</span>
				</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="5">
						Total: ' . $count . ' Seller' . ( $count == 1 ? '' : 's' ) . '
					</td>
					<td>
						<span class="alert alert-info" title="Completed">' . $count_orders_to_pay_out . '</span>
					</td>
				</tr>
			</tfoot>
		</table>';

}


/**
 * Check for sellers that have orders, and calculate any difference between completed orders and payouts.
 */
function bt_payouts_report_audit_imbalance( $args = [] ) {
	// get IDs of all users with orders
	global $wpdb;

	$query = "
SELECT DISTINCT `post_author`
FROM `{$wpdb->prefix}posts`
WHERE `post_type` = 'shop_order' AND `post_status` IN ('wc-completed','wc-processing','wc-refunded','wc-ready-pickup','wc-completed-picked','wc-completed-shipped') 
ORDER BY `post_author`
";
	$user_ids = $wpdb->get_col( $query, 0 );

	// default query arguments for getting all sellers
	$defaults = [
		'post_type' => bt_use_sellers( 'seller_post_type' ),
		'nopaging' => true,
		'posts_per_page' => -1, //$max_rows,
		'post_status' => [ 'published' ],
		'orderby' => 'title',
		'order' => 'ASC',
		'author__in' => $user_ids,
	];

	// exclude any sellers set by bt-sellers theme support
	bt_exclude_sellers_from_reports_arg( $defaults );

	// merge in passed args with defaults
	$parsed_args = wp_parse_args( $args, $defaults );

	// run query
	$sellers = new \WP_Query( $parsed_args );

	// if none, message and exit
	if ( empty( $sellers->posts ) ) {
		echo '<p><em>There aren\'t any sellers to process for this report.</em></p>';
		return;
	}

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>Seller Post</th>
					<th>User</th>
					<th>Email</th>
					<th colspan="2" class="number">Orders Should be Paid</th>
					<th class="number">Diff</th>
					<th colspan="2" class="number">Payouts</th>
				</tr>
			</thead>
			<tbody>';

	// init tracking vars
	$count_sellers = 0;
	$count_should_be_paid_out = 0;
	$count_payouts = 0;
	$dollars_should_be_paid_out = 0;
	$dollars_payouts = 0;
	$dollars_difference = 0;

	// loop sellers
	foreach ( $sellers->posts as $seller ) {
		// set user from author
		$user_id = $seller->post_author;

		// seller data
		$email = get_the_author_meta( 'user_email', $user_id  );
		$platform_fee_rate = bt_seller_processing_fee_rate( $user_id );

		// determine date for which orders must be OLDER THAN in order to pay out
		$payout_orders_through = bt_payout_orders_through();

		// order status counts
		// this is all time, through the most recent eligible payout week
		$seller_orders_should_be_paid_out = bt_get_user_post_count(
			$user_id,
			'shop_order',
			"'wc-completed','wc-completed-picked','wc-completed-shipped'",
			'', // all time
			$payout_orders_through // midnight of Monday of last week if before Wednesday, else Monday of this week
		);

		// skip sellers without orders
		if ( $seller_orders_should_be_paid_out == 0 ) {
			continue;
		}

		// query for user's orders and total up net (total minus platform fee)
		$orders = new \WP_Query( [
			'post_type' => 'shop_order',
			'author' => $user_id,
			'posts_per_page' => -1,
			'post_status' => [
				'wc-completed',
				'wc-completed-picked',
				'wc-completed-shipped',
				// 'wc-refunded', ==> we'll only get completed, as refunds should come out in the wash
			],
			'date_query' => [
				[
					'before' => $payout_orders_through,
					'inclusive' => true,
				],
			],
			'orderby' => 'date',
			'order' => 'ASC',
		] );
		$seller_order_dollars = 0;
		foreach ( $orders->posts as $order ) {
			// get order total from post meta so we don't have to deal with objects unless necessary
			$order_total = get_post_meta( $order->ID, '_order_total', true );

			// check for a partial refund, by looking for a stripe refund transaction id
			if ( get_post_meta( $order->ID, '_stripe_refund_id', true ) ) {
				$order_object = new \WC_Order( $order->ID );
				$order_total -= (float) $order_object->get_total_refunded();
			}

			// determine net after platform fee
			$order_net = $order_total - round( $order_total * $platform_fee_rate, 2 );
			//error_log( __METHOD__ . " -- user_id = $user_id // order_total = $order_total // order_net = $order_net // platform_fee_rate = $platform_fee_rate" );
			$seller_order_dollars += $order_net;
		}

		// get completed payouts
		$seller_payouts_completed = bt_seller_payouts( $user_id );
		$seller_payouts_count = count( $seller_payouts_completed );
		$seller_payouts_dollars = 0;
		//error_log( __METHOD__ . ' -- payouts = ' . print_r( $seller_payouts_completed, true ) );
		foreach ( $seller_payouts_completed as $payout ) {
			$seller_payouts_dollars += $payout->amount;
		}

		// calculate difference
		$seller_difference = round( $seller_payouts_dollars - $seller_order_dollars, 2 );
		$difference_class = 'success';
		if ( $seller_difference < 0 ) {
			// haven't paid out as much as we should have!
			$difference_class = 'danger';
		} else if ( $seller_difference > 0 ) {
			// paid out too much!
			$difference_class = 'warning';
		}

		// count up
		$count_sellers++;
		$count_should_be_paid_out += $seller_orders_should_be_paid_out;
		$count_payouts += $seller_payouts_count;
		$dollars_should_be_paid_out += $seller_order_dollars;
		$dollars_payouts += $seller_payouts_dollars;
		$dollars_difference += $seller_difference;

		// output row for seller
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $seller->ID . '&action=edit">' . $seller->post_title . '</a>
					(' . ucfirst( $seller->post_type ) . ')
				</td>
				<td>
					<a href="' . get_admin_url() . 'user-edit.php?user_id=' . $user_id . '">User ' . $user_id . '</a>
				</td>
				<td>
					<a href="mailto:' . $email . '">' . $email . '</a>
				</td>
				<td class="number">
					<span class="alert alert-info" title="Completed">
						<a href="' . get_admin_url() . 'edit.php?post_type=shop_order&author=' . $user_id . '">' . $seller_orders_should_be_paid_out . '</a>
					</span>
				</td>
				<td class="number">$' . number_format( $seller_order_dollars, 2 ) . '</td>
				<td class="number">
					<span class="alert alert-' . $difference_class . '">$' . number_format( $seller_difference, 2 ) . '</span></td>
				<td class="number">$' . number_format( $seller_payouts_dollars, 2 ) . '</td>
				<td class="number">' . $seller_payouts_count . '</td>
			</tr>
		';
	}

	// final difference class
	$difference_class = 'success';
	if ( $dollars_difference < 0 ) {
		// haven't paid out as much as we should have!
		$difference_class = 'danger';
	} else if ( $dollars_difference > 0 ) {
		// paid out too much!
		$difference_class = 'warning';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="3">
						Total: ' . $count_sellers . ' Seller' . ( $count_sellers == 1 ? '' : 's' ) . ' with Orders
					</td>
					<td>
						<span class="alert alert-info" title="Completed">' . $count_should_be_paid_out . '</span>
					</td>
					<td class="number">$' . number_format( $dollars_should_be_paid_out, 2 ) . '</td>
					<td class="number">
						<span class="alert alert-' . $difference_class . '">$' . number_format( $dollars_difference, 2 ) . '</span>
					</td>
					<td class="number">$' . number_format( $dollars_payouts, 2 ) . '</td>
					<td class="number">' . $count_payouts . '</td>
				</tr>
			</tfoot>
		</table>';

}


/**
 * Calculate the placed date/time through which orders can be paid out.
 *
 * @throws Exception
 * @return string Date formatted as Y-m-d H:i:s
 */
function bt_payout_orders_through() {
	$current_day_of_week = wp_date( 'N' );
	//error_log( __METHOD__ . ': current_day_of_week = ' . $current_day_of_week );

	// determine date for which orders must be OLDER THAN in order to pay out
	if ( $current_day_of_week >= 3 ) {
		// Wednesday or later in week - we can pay out last week, not this week
		$payout_orders_through_datetime = new \DateTime( 'monday this week', new \DateTimeZone( get_option( 'timezone_string' ) ) );
		$payout_orders_through = $payout_orders_through_datetime->format( 'Y-m-d H:i:s' );
		//error_log( __METHOD__ . ': Wednesday or later, orders through = ' . $payout_orders_through );
	} else {
		// before Wednesday - we can pay out the week before last, not last week
		$payout_orders_through_datetime = new \DateTime( 'monday last week', new \DateTimeZone( get_option( 'timezone_string' ) ) );
		$payout_orders_through = $payout_orders_through_datetime->format( 'Y-m-d H:i:s' );
		//error_log( __METHOD__ . ': before Wednesday, orders through = ' . $payout_orders_through );
	}

	return $payout_orders_through;
}


/**
 * Display all payouts made.
 */
function bt_payouts_report_historical() {
	global $wpdb;

	$query = "
SELECT 
	`payouts`.*,
	`users`.`display_name` AS `seller`
FROM `{$wpdb->prefix}bt_payouts` `payouts`
	LEFT JOIN `{$wpdb->prefix}users` `users` ON `users`.`ID` = `payouts`.`user_id`
ORDER BY `date_paid` DESC
";

	$results = $wpdb->get_results( $query );

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>Date</th>
					<th>Seller</th>
					<th>Amount</th>
					<th>Adjustment</th>
					<th>Rate</th>
					<th>Stripe Transfer ID</th>
				</tr>
			</thead>
			<tbody>';

	// loop results
	$count = 0;
	$total_amount = 0;
	foreach ( $results as $result ) {
		$count++;
		$total_amount += $result->amount;
		$total_amount += $result->adjustment;
		// output row for order/customer
		echo '
			<tr>
				<td>' . $result->date_paid . '</td>
				<td>' . $result->seller . '</td>
				<td>$' . number_format( $result->amount, 2 ) . '</td>
				<td>$' . number_format( $result->adjustment, 2 ) . '</td>
				<td>' . ( 100 * $result->fee_rate ) . '%</td>
				<td>' . $result->stripe_transfer_id . '</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="6">
						Total: $' . ( number_format( $total_amount, 2 ) ) . ' over ' . $count . ' payouts.
					</td>
				</tr>
			</tfoot>
		</table>';

	//echo '<pre>' . print_r( $results, 1 ) . '</pre>';
}


/**
 * Generate simple list of customers from order data, so sellers can pull email addresses easily.
 *
 * @param int|null $user_id The user to get orders for. Will default to current user if it's a seller.
 */
function bt_seller_customer_list( $user_id = null ) {
	if ( ! is_numeric( $user_id ) && bt_is_seller() ) {
		$user_id = get_current_user_id();
	}

	if ( ! is_numeric( $user_id ) ) {
		// shouldn't happen, but just in case
		wp_die();
	}

	global $wpdb;

	$query = "
SELECT 
	p.ID, 
	p.post_date as `placed`,
	meta__billing_email.meta_value AS `email`,
	meta__billing_first_name.meta_value AS `firstname`,
	meta__billing_last_name.meta_value AS `lastname`,
	p.post_excerpt as `comments`
FROM `{$wpdb->prefix}posts` `p`
	LEFT JOIN `{$wpdb->prefix}postmeta` meta__billing_email ON meta__billing_email.post_id = p.ID
	LEFT JOIN `{$wpdb->prefix}postmeta` meta__billing_first_name ON meta__billing_first_name.post_id = p.ID
	LEFT JOIN `{$wpdb->prefix}postmeta` meta__billing_last_name ON meta__billing_last_name.post_id = p.ID
WHERE 1=1
	AND `p`.post_author = {$user_id}
    AND `p`.post_type = 'shop_order'
    AND (`p`.post_status IN ('wc-completed','wc-processing','wc-refunded','wc-ready-pickup','wc-completed-picked','wc-completed-shipped'))
   	AND meta__billing_email.meta_key = '_billing_email'
   	AND meta__billing_first_name.meta_key = '_billing_first_name'
   	AND meta__billing_last_name.meta_key = '_billing_last_name'
ORDER BY p.post_date DESC
";

	$results = $wpdb->get_results( $query );

	// start report table
	echo '
		<table class="bt-report">
			<thead>
				<tr>
					<th>Order ID</th>
					<th>Order Placed</th>
					<th>Email</th>
					<th>First Name</th>
					<th>Last Name</th>
					<th>Comments</th>
				</tr>
			</thead>
			<tbody>';

	// loop results
	$count = 0;
	foreach ( $results as $result ) {
		$count++;
		// output row for order/customer
		echo '
			<tr>
				<td>
					<a href="' . get_admin_url() . 'post.php?post=' . $result->ID . '&action=edit">' . $result->ID . '</a>
				</td>
				<td>' . $result->placed . '</td>
				<td>
					<a href="mailto:' . $result->email . '">' . $result->email . '</a>
				</td>
				<td>' . $result->firstname . '</td>
				<td>' . $result->lastname . '</td>
				<td>' . $result->comments . '</td>
			</tr>
		';
	}

	// finish table
	echo '
			</tbody>
			<tfoot>
				<tr>
					<td colspan="6">
						Total: ' . $count . '
					</td>
				</tr>
			</tfoot>
		</table>';

	//echo '<pre>' . print_r( $results, 1 ) . '</pre>';
}


/**
 * @param int $user_id
 *
 * @return string|null
 */
function bt_get_seller_oldest_open_order_date( $user_id ) {
	if ( ! is_numeric( $user_id ) ) {
		wp_die( __METHOD__ . ': no user ID provided.' );
	}

	$post_status = "'wc-processing','wc-ready-pickup'";

	global $wpdb;

	$query = "
            SELECT post_date
            FROM {$wpdb->posts}
            WHERE 1=1
            	AND {$wpdb->posts}.post_author = $user_id
                AND {$wpdb->posts}.post_type = 'shop_order'
                AND ({$wpdb->posts}.post_status IN ($post_status))
            ORDER BY post_date ASC
            LIMIT 1
            ";

	//error_log( 'SELLER OLDEST ORDER: query = ' . $query );

	return $wpdb->get_var( $query );

}


/**
 * Check items on an order to see if it has multiple sellers.
 *
 * @param \WC_Order $order
 *
 * @return bool
 */
function bt_order_has_multiple_sellers( $order ) {
	$seller_user_id = null;
	foreach ( $order->get_items() as $item ) {
		//error_log( __METHOD__ . ': item = ' . $item->get_product_id() . ' / ' . $item->get_name() /*print_r( $item, 1 )*/ );

		// get seller user id for this item
		$item_user_id = bt_get_author_id_for_post( $item->get_product_id() );

		// track seller if we haven't yet
		if ( empty( $seller_user_id ) ) {
			$seller_user_id = $item_user_id;
		} else if ( $item_user_id != $seller_user_id ) {
			// seller has been set previously and is different from this item!
			return true;
		}
	}

	return false;
}


/**
 * Add a 'post__not_in' arg if the site has sellers to exclude, like NEMAA for cml/nma.
 *
 * @param $args
 */
function bt_exclude_sellers_from_reports_arg( &$args ) {
	//error_log( __METHOD__ . ' -- args = ' . print_r( $args, true ) );
	if ( is_array( $exclude_sellers = bt_use_sellers( 'sellers_exclude_from_reports' ) ) ) {
		$args['post__not_in'] = $exclude_sellers;
		//error_log( __METHOD__ . ' -- UPDATED args = ' . print_r( $args, true ) );
	}
}

/*
 * wc_admin_reports_path - can filter this to change where we include stuff from, like custom tips report
 *
 *

function bt_filter_wc_report_args( $args ) {
	error_log( 'FILTER REPORT ARGS: args = ' . print_r( $args, 1 ) );

	return $args;
}
//add_filter( 'woocommerce_reports_get_order_report_data_args', 'bt_filter_wc_report_args', 9988, 1 );


 * woocommerce_admin_report_data - for 3P filtering; what's in here?? => OLD REPORTS, ALL DATA
 * woocommerce_admin_report_chart_data - same!
 */
